(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
	typeof define === 'function' && define.amd ? define(factory) :
	(global.topologica = factory());
}(this, (function () { 'use strict';

var Graph = () => {
  const edges = {};

  const adjacent = node => edges[node] || [];

  const addNode = node => {
    edges[node] = adjacent(node);
  };

  const addEdge = (u, v) => {
    addNode(u);
    addNode(v);
    adjacent(u).push(v);
  };

  const depthFirstSearch = sourceNodes => {
    const visited = new Set();
    const nodeList = [];

    const DFSVisit = node => {
      if (!visited.has(node)) {
        visited.add(node);
        adjacent(node).forEach(DFSVisit);
        nodeList.push(node);
      }
    };

    sourceNodes.forEach(node => {
      visited.add(node);
    });

    sourceNodes.forEach(node => {
      adjacent(node).forEach(DFSVisit);
    });

    return nodeList;
  };

  const topologicalSort = sourceNodes => {
    return depthFirstSearch(sourceNodes).reverse();
  };

  return { addEdge, topologicalSort };
}

var isReactiveFunction = value => (
  Array.isArray(value)
  && value.length === 2
  && typeof value[0] === 'function'
  && typeof value[1] === 'string'
)

const parse = inputsStr => inputsStr
  .split(',')
  .map(input => input.trim());

var unpackReactiveFunction = ([fn, inputsStr]) => ({
  fn,
  inputs: parse(inputsStr)
})

var main = options => {
  const values = new Map();
  const changed = new Set();
  const functions = new Map();
  const graph = Graph();

  const digest = () => {
    const changedValues = Array.from(changed.values());
    graph
      .topologicalSort(changedValues)
      .forEach(property => functions.get(property)());
    changed.clear();
  };

  const setProperty = ([property, value]) => {
    values.set(property, value);
    changed.add(property);
  };

  const set = options => {
    Object.entries(options).forEach(setProperty);
    digest();
  };

  const get = property => values.get(property);

  const getAll = properties => properties.reduce((obj, property) => {
    obj[property] = get(property);
    return obj;
  }, {});

  Object.entries(options).forEach(entry => {
    const [property, value] = entry;
    if (isReactiveFunction(value)) {
      const { fn, inputs } = unpackReactiveFunction(value);
      inputs.forEach(input => graph.addEdge(input, property));
      functions.set(property, () => {
        values.set(property, fn(getAll(inputs)));
      });
    } else {
      functions.set(property, () => get(property));
      setProperty(entry);
    }
  });

  digest();

  return { set, get };
};

return main;

})));
