(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
	typeof define === 'function' && define.amd ? define(factory) :
	(global.topologica = factory());
}(this, (function () { 'use strict';

var Graph = (function () {
  var edges = {};

  var adjacent = function adjacent(node) {
    return edges[node] || [];
  };

  var addNode = function addNode(node) {
    edges[node] = adjacent(node);
  };

  var addEdge = function addEdge(u, v) {
    addNode(u);
    addNode(v);
    adjacent(u).push(v);
  };

  var depthFirstSearch = function depthFirstSearch(sourceNodes) {
    var visited = new Set();
    var nodeList = [];

    var DFSVisit = function DFSVisit(node) {
      if (!visited.has(node)) {
        visited.add(node);
        adjacent(node).forEach(DFSVisit);
        nodeList.push(node);
      }
    };

    sourceNodes.forEach(function (node) {
      visited.add(node);
    });

    sourceNodes.forEach(function (node) {
      adjacent(node).forEach(DFSVisit);
    });

    return nodeList;
  };

  var topologicalSort = function topologicalSort(sourceNodes) {
    return depthFirstSearch(sourceNodes).reverse();
  };

  return { addEdge: addEdge, topologicalSort: topologicalSort };
});

var isReactiveFunction = (function (value) {
  return Array.isArray(value) && value.length === 2 && typeof value[0] === 'function' && typeof value[1] === 'string';
});

var slicedToArray = function () {
  function sliceIterator(arr, i) {
    var _arr = [];
    var _n = true;
    var _d = false;
    var _e = undefined;

    try {
      for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
        _arr.push(_s.value);

        if (i && _arr.length === i) break;
      }
    } catch (err) {
      _d = true;
      _e = err;
    } finally {
      try {
        if (!_n && _i["return"]) _i["return"]();
      } finally {
        if (_d) throw _e;
      }
    }

    return _arr;
  }

  return function (arr, i) {
    if (Array.isArray(arr)) {
      return arr;
    } else if (Symbol.iterator in Object(arr)) {
      return sliceIterator(arr, i);
    } else {
      throw new TypeError("Invalid attempt to destructure non-iterable instance");
    }
  };
}();

var parse = function parse(inputsStr) {
  return inputsStr.split(',').map(function (input) {
    return input.trim();
  });
};

var unpackReactiveFunction = (function (_ref) {
  var _ref2 = slicedToArray(_ref, 2),
      fn = _ref2[0],
      inputsStr = _ref2[1];

  return {
    fn: fn,
    inputs: parse(inputsStr)
  };
});

var main = (function (options) {
  var values = new Map();
  var changed = new Set();
  var functions = new Map();
  var graph = Graph();

  var digest = function digest() {
    graph.topologicalSort(Array.from(changed.values())).forEach(function (property) {
      return functions.get(property)();
    });
    changed.clear();
  };

  var setProperty = function setProperty(_ref) {
    var _ref2 = slicedToArray(_ref, 2),
        property = _ref2[0],
        value = _ref2[1];

    values.set(property, value);
    changed.add(property);
  };

  var set$$1 = function set$$1(options) {
    Object.entries(options).forEach(setProperty);
    digest();
  };

  var get$$1 = function get$$1(property) {
    return values.get(property);
  };

  var getAll = function getAll(properties) {
    return properties.reduce(function (obj, property) {
      obj[property] = get$$1(property);
      return obj;
    }, {});
  };

  var allDefined = function allDefined(properties) {
    return properties.every(function (property) {
      return values.get(property);
    });
  };

  Object.entries(options).forEach(function (entry) {
    var _entry = slicedToArray(entry, 2),
        property = _entry[0],
        value = _entry[1];

    if (isReactiveFunction(value)) {
      var _unpackReactiveFuncti = unpackReactiveFunction(value),
          fn = _unpackReactiveFuncti.fn,
          inputs = _unpackReactiveFuncti.inputs;

      inputs.forEach(function (input) {
        graph.addEdge(input, property);
      });

      functions.set(property, function () {
        if (allDefined(inputs)) {
          values.set(property, fn(getAll(inputs)));
        }
      });
    } else {
      functions.set(property, function () {
        return get$$1(property);
      });
      setProperty(entry);
    }
  });

  digest();

  return { set: set$$1, get: get$$1 };
});

return main;

})));
